#include "Ball.h"
#include "globals.h"
#include <QtOpenGL>
#include "Game.h"

extern Game g;

// constructor
Ball::Ball() : moving(false), previousSpeed() {}

// constructor
Ball::Ball(const Point & pcenter, float pradius, float pweight, const QColor & pcolor, int pnumber) : center(pcenter), radius(pradius), weight(pweight), color(pcolor), speed(0,0,0), number(pnumber), moving(false), previousSpeed() {
}

// distance from another ball
float Ball::distance(const Ball & ball) const {
	Direction connectLink = Direction(center, ball.getCenter());
	return connectLink.size();
}

// move using speed vector
void Ball::move(float f=1.0) {
	if (f==1) {
		lastIterationPosition = center;		// save position at the beginning of each stroke
	}
	previousCenter = center;
	center = center + speed*f;
	if (!speed.isNull()) {
		moving = true;
	}
}

Point Ball::getCenter() const {
	return center;
}

Direction Ball::getSpeed() const {
	return speed;
}

float Ball::getRadius() const {
	return radius;
}

float Ball::getWeight() const {
	return weight;
}

void Ball::setSpeed(const Direction & newSpeed) {
	previousSpeed = speed;
	speed = newSpeed;
}

void Ball::setCenter(const Point& p) {
	previousCenter = center;
	center = p;
}

int Ball::getNumber() const {
	return number;
}

QColor Ball::getColor() const {
	return color;
}

// draw itself on screen - OpenGL
void Ball::draw() const {
	glColor3f((float)color.red()/255, (float)color.green()/255, (float)color.blue()/255);
	glMatrixMode( GL_MODELVIEW );
	glPushMatrix();
	glLoadIdentity();
	glTranslated( center.getX(), center.getY(), center.getZ() );

	GLUquadricObj* q = gluNewQuadric();
	gluQuadricOrientation( q, GLU_OUTSIDE );
    gluQuadricDrawStyle( q, GLU_FILL );
    gluQuadricNormals( q, GLU_SMOOTH );
    gluSphere( q, radius, 30, 30 );
    gluDeleteQuadric( q );

	glPopMatrix();
}

// collision detection
// balls are colliding if their centers are closer than radius1 + radius2
bool Ball::colliding(const Ball& b) const {
	Direction axis = b.getCenter() - center;
	return ( axis.size() < (b.getRadius() + radius) );
}

bool Ball::isWhite() const {
	return (number==0);
}

bool Ball::isBlack() const {
	return (number==8);
}

bool Ball::isSolid() const {
	return (number >= 1 && number <= 7);
}

bool Ball::isStripe() const {
	return (number >= 9 && number <= 15);
}

// position in the last moment (it's different from lastIterationPosition)
Ball Ball::previousPosition() const {
	Ball b = *this;
	b.setCenter(center + (-speed));
	return b;
}

// compute how much are the balls intersecting
// returns 0<t<1 saying how much of the last move needs to be taken back to get the balls in touching position
// solving a quadratic equation
float Ball::computeOverlap(const Ball& b) const {
	float D = radius + b.radius;

	// center2 - center1
	float A = b.center.getX() - center.getX();
	float B = b.center.getY() - center.getY();
	float H = b.center.getZ() - center.getZ();

	// speed2 - speed1
	float E = b.speed.getX() - speed.getX();
	float F = b.speed.getY() - speed.getY();
	float G = b.speed.getZ() - speed.getZ();

	float p = E*E + F*F + G*G;
	float q = 2 * (A*E + B*F + H*G);
	float r = A*A + B*B + H*H - D*D;

	if (p==0) {		// if balls move at the same speed, don't do anything
		return 0;
	}

	assert(q*q - 4*p*r >= 0);

	float dis = sqrt(q*q - 4*p*r);
	float t = -abs( (-q - dis) / (2*p) );	// the other root "t = (-q + dis) / (2*p);" is more likely to be more than 0 and as such not suitable
	float t1 = -abs( (-q + dis) / (2*p) );
	if (t1 > t) {	// absolute t1 is less than absolute t
		t = t1;
	}
	
	// safety corrections to eliminate round off errors etc.
	if (t <= -1) {
		t = -0.97;
	}
	if (t > 0) {
		t = 0;
	}
	assert(t>=-1 && t<=0);

	return t;
}

// decrease speed of the ball
// used in each iteration and upon each collision
void Ball::attenuate(float f) {
	if (speed.size() < 10*MIN_SPEED) {		// for low speeds use a stronger attenuation factor
		speed = speed * (f*f);
	}
	else {		// normal situation
		speed = speed * f;
	}
	if (speed.size() < MIN_SPEED) {		// if speed is small enough stop the ball for good
		speed = Direction();
		moving = false;
	}
}

// is the ball moving (i.e. its speed vector is not zero)
bool Ball::isMoving() const {
	return moving;
}

// ball was stationary when the last stroke took place
bool Ball::wasStanding() const {
	return previousSpeed.isNull();
}

// position of the ball when the last stroke took place (differs from lastPosition)
Point Ball::getLastIterationPosition() const {
	return lastIterationPosition;
}

// ball out of bounds of the playing area
// assumptions:
//	- center of the table is in 0,0
//	- table is symetric
//	- table is horizontal
bool Ball::isInPocket() const {
	return (abs(center.getX()) > TABLE_LENGTH/2 || abs(center.getY()) > TABLE_WIDTH/2);		
}

// equals operator
bool Ball::operator==(const Ball& b) const {
	return (number == b.number && center == b.center && speed == b.speed);
}