#include "Collision.h"

// constructor
Collision::Collision() : firstObject(), secondObject(), t(0) {}

// constructor
Collision::Collision(Ball *b, Plane &p, float d) : firstObject(b), plane(p), t(d) {
	secondObject = (Plane*) &plane;
}

// constructor
Collision::Collision(Ball *b1, Ball *b2, float d) : firstObject(b1), t(d) {
	secondObject = (Ball*) b2;
}

// copy constructor
Collision::Collision(const Collision& c) : t(c.t), firstObject(c.firstObject), secondObject(c.secondObject), plane(c.plane) {
	if (c.secondObject == &c.plane) {
		secondObject = (Plane*) &plane;
	}
}

// used only when talking about ball-wall collision
// if empty it means that there was no collision
bool Collision::isEmpty() const {
	return plane.isNull();
}

// solve the collision
// compute new speed vectors for the ball(s)
// the two types of collision are solved in diferrent ways
void Collision::handle() {
	
	// ball - cuboid collision
	// algorithm: mirror the speed vector with respect to normal of the plane
	if (secondObject == &plane) {
		Ball *ball = firstObject;	// just for easier access
		Plane *pl = (Plane*) secondObject; // just for easier access

		Point x1 = ball->getCenter() + (-pl->getNormal() * 2 * pl->distance(*ball));	// mirrored ball center
		Point p = pl->findIntersection(ball->getCenter(), ball->getSpeed());	// intersection of direction and colliding plane
		ball->setSpeed((p-x1) * (ATTENUATION_BALLWALL*ball->getSpeed().size() / (p-x1).size()));	// set new speed, use impact attenuation
	}

	
	// ball - ball collision
	// algorithm: compute axis between ball centers, do a projection of speed vector with respect of the axis
	//			  the component that lies on the axis is important, the other will stay the say after the impact
	//			  use inertia and energy equations to solve it
	else {
		Ball *b1 = firstObject;		// just for easier access
		Ball *b2 = (Ball*) secondObject;	// just for easier access
		
		Direction connectingAxis = b1->getCenter() - b2->getCenter();
		connectingAxis.normalize();

		float tmp = (connectingAxis * b1->getSpeed());
		Direction U1x = connectingAxis * tmp;	// projection of speed of ball1 on the axis
		Direction U1y = b1->getSpeed() - U1x;	// perpendicular remainder
		tmp = -connectingAxis * b2->getSpeed();
		Direction U2x = -connectingAxis * tmp;	// projection of speed of ball2 on the axis
		Direction U2y = b2->getSpeed() - U2x;	// perpendicular remainder

		Direction V1x, V2x;
		if ( b1->getWeight() == b2->getWeight() ) {	// balls have the same weight - just switch the velocities
			V1x = U2x;
			V2x = U1x;
		}
		else {	// balls with different weights (cue ball is slightly heavier) - compute new speeds
			tmp = b1->getWeight() + b2->getWeight();		
			V1x = ((U1x*b1->getWeight()) + (U2x*b2->getWeight()) - (U1x-U2x)*b2->getWeight()) / tmp;
			V2x = ((U1x*b1->getWeight()) + (U2x*b2->getWeight()) - (U2x-U1x)*b1->getWeight()) / tmp;
		}
		Direction V1 = V1x + U1y;	// ball1 new speed
		Direction V2 = V2x + U2y;	// the other ball's new speed
		
		float attenuation = ATTENUATION_BALLBALL;
				
		// when balls in touching position, the force is transfered without losses, therefore no attenuation
		if ((b1->getSpeed().isNull() && b2->wasStanding()) || (b1->wasStanding() && b2->getSpeed().isNull())) {
			float distance = b1->distance(*b2);
			float previousDistance = (b1->getLastIterationPosition() - b2->getLastIterationPosition()).size();
			if (previousDistance > 0.98*distance && previousDistance < 1.02*distance) {		// approximation of touching position (distance should be approx. equal to 2*radius)
				attenuation = 1.1;
			}
		}

		// set new speeds for the balls, attenuation factor is used
		b1->setSpeed(V1*attenuation);
		b2->setSpeed(V2*attenuation);
	}
}

// assignment operator
Collision& Collision::operator=(const Collision& c) {
	t = c.t;
	plane = c.plane;
	firstObject = c.firstObject;
	if (c.secondObject == &c.plane) {	// ball - cuboid collision ... change pointer to point to the new instance of the plane
		secondObject = (Plane*) &plane;
	}
	else {	// ball - ball collision ... pointing to the same ball as the source collision
		secondObject = c.secondObject;
	}
	return *this;
}

// getter for the t factor
float Collision::getT() const {
	return t;
}