// cuboid is like a box - 6 perpendicular sides

#include "Cuboid.h"
#include "Point.h"
#include "Direction.h"
#include "Ball.h"

// constructor
Cuboid::Cuboid() {}

// constructor
Cuboid::Cuboid(const Point & ppoint, const Direction & dir1, const Direction & dir2, const Direction & dir3, const QColor & color) {
	// create walls (be careful with the order of direction vectors - important for correct normals computation)
	walls[0] = Wall(ppoint, dir1, dir3, color);
	walls[1] = Wall(ppoint+dir1, dir2, dir3, color);
	walls[2] = Wall(ppoint+dir2, dir3, dir1, color);
	walls[3] = Wall(ppoint, dir3, dir2, color);
	walls[4] = Wall(ppoint, dir2, dir1, color);
	walls[5] = Wall(ppoint+dir3, dir1, dir2, color);
	Direction diameter = (ppoint + dir1 + dir2 + dir3) - ppoint;
	radius = diameter.size() / 2;		// radius of the bounding sphere
	center = ppoint + (diameter/2);		// central point of the cuboid
}

// given a ball, find a colliding plane
// it can be a plane which one of the walls lie on
// or it can be a computed plane (constructed using two or three walls)
// ball hits a wall somewhere in the middle -> return a wall plane
// ball hits an edge -> return a plane that is 45 to both walls and their intersection lies in the plane
// ball hits a corner -> return a plane that is 45 to all three walls
Collision Cuboid::findCollisionPlane(Ball *ball) const throw(insideCuboidException) {

	// compute distance for each wall of the cuboid (handle them as plains for now)
	float distance[6];
	int colliding = 0, negative = 0, positiveNoCollision = 0;
	int collidingWalls[4];	// three is enough, four is safer
	Collision result;

	if (ball->getSpeed().isNull()) {	// the ball is partly in the cuboid, but it's not moving - normally this shall never happen, but it is good to have it here
		return result;					// e.g. upon incorrect set up of the world
	}

	for (int i=0; i<6; i++) {
		distance[i] = walls[i].distance(*ball);
		float previousDistance = walls[i].distance(ball->previousPosition());
		// if there has been a collision with the plane
		if (distance[i] >= 0 && distance[i] <= ball->getRadius()) {
			if (previousDistance >= 0) {	// collision with the plane
				collidingWalls[colliding] = i;
				++colliding;
				assert(colliding < 4);	// colliding >= 4 doesn't make sense
			}
			else {	// ball came from behind the plane (not from behind the wall, but from next to it, naturallz), there cannot be any collision
				++positiveNoCollision;
			}
		}
		else if (distance[i] < 0) {
			++negative;
		}
	}

	if (negative >= 6) {
		throw insideCuboidException(ball);	// if negative == 6, the ball is inside the cuboid
	}

	// one colliding wall - either simple rebound from a wall or no rebound at all
	if (colliding == 1) {
		// if one is colliding and at least one is positive, the ball is too far - there has been no collision
		if (!(negative < 5 && positiveNoCollision == 0)) {
			// compute touching point - using similar triangles
			// |C-I| / d == |X-I| / R
			// C .. ball center initial, I .. intersection, X .. ball center final, d .. distance initial, r .. radius (i.e.distance final)
			// t = (|X-I| - |C-I|) / |speed|
			Point intersection = walls[collidingWalls[0]].findIntersection(ball->getCenter(), ball->getSpeed());
			float t = (ball->getCenter() - intersection).size() * (ball->getRadius() / distance[collidingWalls[0]] - 1) / ball->getSpeed().size();
			
			// eliminate computation errors
			if (t>=1) {		// may occur, e.g. when there's big difference in x and y speed sizes and ball near corner - this is safe and doesn't affect the result much
				t = 0.98;
			}
			if (t<0) {
				t = 0;
			}
			if (!(t<1 && t>=0)) {
				assert(t<1 && t>=0);
			}
			
			result = Collision(ball,(Plane) walls[collidingWalls[0]], -t);	// returns a plane that the wall lies on
		}
	}

	// two colliding walls - edge rebound
	else if (colliding == 2) {
		// if two are colliding and at least one is positive, the ball is too far - there has been no collision
		if (negative < 4 && positiveNoCollision == 0) {
			// nothing to be done here, result remains empty
		}
		else {
			WallPoints points1 = walls[ collidingWalls[0] ].getPoints();
			WallPoints points2 = walls[ collidingWalls[1] ].getPoints();
			bool found = false, found1 = false;
			Point testPoint, testPoint1;

			// find points that belong to both walls
			for (int i=0; i<4; ++i) {
				if (!found) {
					testPoint = points1[i];
				}
				testPoint1 = points1[i];
				for (int j=0; j<4; ++j) {
					if (!found && testPoint == points2[j]) {
						found = true;
						continue;
					}
					else if (testPoint1 == points2[j]) {
						found1 = true;
						break;
					}
				}
				if (found1) {
					break;
				}
			}

			assert(found1);	// if not found, it means that the two walls are not neighbors - error
			
			// compute distance between the line formed by intersection of the two walls and the center of the ball
			// (it could be close enough to both of the wall planes, but not close enough to the corner)
			// this should ideally equal to radius1+radius2 when in touching point, but computing it would be too complicated
			// instead we compute touching point to the newPlane
			Direction r0 = Direction(Point(), testPoint);
			Direction s = testPoint - testPoint1;
			Direction r1 = Direction(Point(), ball->getCenter());

			Direction dist = r0 - r1 + s* ((r1-r0)*s/(s*s));
			float trueDistance = dist.size();

			// there has really been a collision
			if (trueDistance < ball->getRadius()) {	
				// compute an "edge plane" (45 degrees to each of the two planes)
				// normal = normal1 + normal2
				Plane newPlane = Plane(testPoint, walls[ collidingWalls[0] ].getNormal() + walls[ collidingWalls[1] ].getNormal());
				
				// compute distance again (it could be close enough to both of the walls, but not close enough to the corner plane)
				float distance = newPlane.distance(*ball);

				// compute touching point - using similar triangles
				// |C-I| / d == |X-I| / R
				// C .. ball center initial, I .. intersection, X .. ball center final, d .. distance initial, r .. radius (i.e.distance final)
				// t = (|X-I| - |C-I|) / |speed|
				Point intersection = newPlane.findIntersection(ball->getCenter(), ball->getSpeed());
				float t = (ball->getCenter() - intersection).size() * (ball->getRadius() / distance - 1) / ball->getSpeed().size();
				
				// do a correction to minimize the approximation error
				t *= distance / trueDistance;
				if (t>=1) {
					t = 0.97;
				}
				if (t<0) {
					t = 0;
				}
				assert(t<1 && t>=0);

				result = Collision(ball, newPlane, -t);		// returns the new plane (45 to both of the planes)
			}
		}
	}

	// three colliding walls - corner rebound
	// would be used in 3D - not really implemented yet
	else if (colliding == 3) {

		WallPoints points1 = walls[ collidingWalls[0] ].getPoints();
		WallPoints points2 = walls[ collidingWalls[1] ].getPoints();
		WallPoints points3 = walls[ collidingWalls[2] ].getPoints();
		Point testPoint(0,0,0);
		bool found = false;

		// find a point that belongs to all three walls
		for (int i=0; i<4; ++i) {
			testPoint = points1[i];
			
			for (int j=0; j<4; ++j) {
				if (testPoint == points2[j]) {
			
					for (int k=0; k<4; ++k) {
						if (testPoint == points3[k]) {
							found = true;
							break;
						}
					}

				}
				if (found) {
					break;
				}	
			}
			
			if (found) {
				break;
			}
		}

		assert(found);	// if not found, it means that the two walls are not neighbors - error

		// compute a "corner plane" (45 degrees to all of the three planes)
		// normal = normal1 + normal2 + normal3
		Plane newPlane = Plane(testPoint, walls[ collidingWalls[0] ].getNormal() + walls[ collidingWalls[1] ].getNormal() + walls[ collidingWalls[2] ].getNormal());
		
		// compute distance again (it could be close enough to both of the walls, but not close enough to the corner)
		// this is an approximation - see colliding=2 scenario to find more about it
		float distance = newPlane.distance(*ball);
		if (distance < ball->getRadius()) {
			// compute touching point - using similar triangles
			// |C-I| / d == |X-I| / R
			// C .. ball center initial, I .. intersection, X .. ball center final, d .. distance initial, r .. radius (i.e.distance final)
			// t = (|X-I| - |C-I|) / |speed|
			Point intersection = newPlane.findIntersection(ball->getCenter(), ball->getSpeed());
			float t = (ball->getCenter() - intersection).size() * (ball->getRadius() / distance - 1) / ball->getSpeed().size();
			if (t>=1) {
				t = 0.97;
			}
			if (t<0) {
				t = 0;
			}
			assert(t<1 && t>=0);

			result = Collision(ball, newPlane, -t);
		}
	}

	// if no collision detected, return a null plane (implicit)

	return result;
}

// draw itself on the screen 
void Cuboid::draw() const {
	for (int i=0; i<6; ++i) {
		walls[i].draw();
	}
}

// simplest collision detection - using a bounding sphere
// when distance between centers of ball and cuboid is more than sum of the two radii, there couldn't be a collision
bool Cuboid::colliding(const Ball& b) const {
	Direction axis = Direction(b.getCenter() - center);
	return ( axis.size() < (radius + b.getRadius()) );
}
