// brain of the application
// contains game rules
// controls ball moves
// looks for collisions calls function that handle them

#include "Game.h"
#include "Plane.h"
#include <vector>
#include "coolaaber.h"
#include <iostream>
#include <string>
#include <sstream>
#include <algorithm>
using namespace std;

// constructor
Game::Game(const string cfgFile) : setup(cfgFile) {
	newGame();
	overheadView = true;
}

// reset game - reposition all balls, reset state variables
void Game::newGame() {

	balls.clear();
	cuboids.clear();

	setup.readFromFile(balls, cuboids);
	
	Ball* white = getWhite();
	if (!white) {
		throw "No white ball found";
	}

	moving = false;
	cue.setDirection(Direction(2.5,0,0));
	cue.setPosition(white->getCenter());	// position the cue
	resetCueBall = false;
	assert(balls.size() >= setup.countOfSolids+setup.countOfStripes + 1);
	state = NEWGAME;
	activePlayer = PLAYING_FIRST;
	glClearColor(0.0f, 0.0f, 0.0f, 1.0f);
}

// returns the cue ball
Ball* Game::getWhite() {
	for (int i=0; i<balls.size(); ++i) {
		if (balls[i].isWhite()) {
			return &(balls[i]);
		}
	}
	return NULL;
}

// do a stroke - make the cue ball move
void Game::stroke() {
	// if not moving, do a stroke
	if (!moving && state != WIN_SOLID && state != WIN_STRIPE) {
		assert(!cue.getDirection().isNull());
		assert(getWhite());
		getWhite()->setSpeed(cue.getDirection());
		moving = true;
		pocketedMine = pocketedOther = false;
		resetCueBall = false;
		widget->setLabel("...moving...");
		widget->start(TIMER);	// start simulation
	}
	
	// if already moving, stop displaying animation, compute the result and display
	else {	
		widget->stop();
		moving = true;
		widget->setLabel("...moving...");
		while (moving) {	// repeat until all balls come to halt
			doIteration();
		}
		widget->updateGL();	// redraw the scene
	}
}

// move cue to new position (upon mouse movement)
// it rotates its direction vector
void Game::rotateCue(float x, float y) {
	cue.setDirection( Direction(cos(x)*cue.getDirection().getX() - sin(x)*cue.getDirection().getY() , sin(x)*cue.getDirection().getX() + cos(x)*cue.getDirection().getY() , 0 ) );
	updateView();
}

// one iteration of simulation
void Game::doIteration() {
	int iter=3;		// do 3 iterations between displaying on screen
	while (moving && iter>0) {
		move(1);	// move all the balls
		handleCollisions();	// find a solve collisions
		
		// removing balls that fell in a pocket and speed attenuation
		int movingBalls = 0;
		for (int i=0; i<balls.size(); ++i) {
			// remove balls that are out of bounds of the playing field
			if (balls[i].isInPocket()) {
				removeBall(i);
				continue;
			}

			// slow down the ball (just general friction)
			balls[i].attenuate(ATTENUATION_GENERAL);
			if (balls[i].isMoving()) {
				++movingBalls;
			}

		}

		// if no ball is moving - stop the simulation
		if (movingBalls == 0 && state != WIN_SOLID && state != WIN_STRIPE) {	
			widget->stop();	// stop the timer
			moving = false;		
			
			// game logic
			if (resetCueBall) {	// someone pocketed the white ball
				balls.push_back(Ball(Point(WHITE_START_POSITION, RADIUS_WHITE), RADIUS_WHITE, WEIGHT_WHITE, COLOR_WHITE, 0));
			}
			if (resetCueBall || pocketedOther || !pocketedMine) {
				switchPlayers();
			}

			displayPlayer();	// whose turn it is
			assert(getWhite());
			cue.setPosition(getWhite()->getCenter());	// move the cue to the position of the white ball
		}
		--iter;
	}

	updateView();	// move the viewers position if necessary
}

// move all the balls in the game
void Game::move(float t) {
	for (int i=0; i<balls.size(); ++i) {
		balls[i].move(t);
	}
}

// find collisions and handle them appropriatelly
// quite sketchy - it is almost impossible to solve it preciselly so some approximations are being used
void Game::handleCollisions() {
	if (!moving) {	// don't do anything if no ball is moving
		return;
	}

	float t, takenBack=0;
	bool repeat = false, repeatWhole = false, ballLast;
	int last_i = -1, last_k = -1, last_j = -1;
	int wholeIterations = 0;
	Collision lastCollision;
	int countWalls=0, countBalls=0;

	do {	// each time you find a collision and handle it, you have to go through the whole collision detection again
		repeatWhole = false;
		takenBack = 0;	// moment that was the furthest in the past so far discovered
		
		try {
			// for each ball check collision with all other balls and with cuboids in the game
			for (int i=0; i<balls.size(); ++i) {
				
				// collisions with other balls
				for (int k=i+1; k<balls.size(); ++k) {
					if (i==last_i && k==last_k) {	// skip the ones that caused collision in last iteration - they already are in the touching position
						continue;
					}
					if (balls[k].colliding(balls[i])) {	// balls intersecting (their centers are closer than R1+R2)

						t = balls[i].computeOverlap(balls[k]);	// how much in the past when they touched (-1..0)
						
						if (t < takenBack) {	// if these two touched earlier than the previously found collision
							if (t <= -1) {
								t = -0.98;
								break;
							}
							takenBack = t;
							ballLast = true;
							++countBalls;
							last_i = i;
							last_k = k;
						}

					}
				}

				// collisions with cuboids
				for (int j=0; j<cuboids.size(); ++j) {
					if (i==last_i && j==last_j) {	// skip the ones that caused collision in last iteration - they already are in the touching position
						continue;
					}

					if (cuboids[j].colliding(balls[i])) {	// simplest collision detection using a bounding sphere
						Collision c = cuboids[j].findCollisionPlane(&balls[i]);	// returns a constructed collision plane or a null plane
						if (!c.isEmpty()) {	// if collision
							
							t = c.getT();	// how much in the past when they touched (-1..0)
							
							if (t < takenBack) {	// if they touched earlier than the previous collision occured
								if (t<= -1) {
									t = -0.98;
									break;
								}
								takenBack = t;
								ballLast = false;
								++countWalls;
								last_i = i;
								last_j = j;
								lastCollision = c;
							}
						}
					}
				}	// end of cycle through cuboids

			}	// end of outer for-cycle (for each ball)
		
		} catch (insideCuboidException e) {
			// there has been an exception 
			// a ball is inside of a cuboid and is moving (possibly it went too fast)
			// solution - remove the ball from the game and the game is ready to continue (not optimal but better than crash)
			for (vector<Ball>::iterator it = balls.begin(); it<balls.end(); ++it) {
				if (*it == *e.ball) {
					if (it->isWhite()) {
						resetCueBall = true;;
					}
					balls.erase(it);
					break;
				}
			}
		}


		// if a collision was detected
		// move back in time (move all objects in reverse direction to the moment), resolve collision and move forward so no time is lost
		if (takenBack < 0) {
			move(takenBack);
			assert(last_i != -1);

			// ball-ball collision
			if (ballLast) {
				assert(last_k != -1);
				float distance = Direction(balls[last_i].getCenter() - balls[last_k].getCenter()).size();
				Collision col = Collision(&balls[last_i], &balls[last_k]);
				col.handle();	// compute new speed vectors
			}

			// ball-wall collision
			else {	
				assert(last_j != -1);
				lastCollision.handle();	// compute new speed vector
			}

			move(-takenBack);	// do the rest of the move, so no time is lost
			
			// reset some supporting variables
			last_i = last_k = last_j = -1;
			takenBack = 0;
			repeat = false;
			repeatWhole = true;	// this means do the collision check again (there might be new collisions)
		}
		
		if (++wholeIterations > 15) {	// safety constrain - preserves freezing in complicated situations
			repeatWhole = false;
		}

	} while (repeatWhole);	// if a collision was found and resolved, do the collision check again (there might be new collisions)
	
}



/* draw everything to the screen */
void Game::draw() const {
	for (int i=0; i < balls.size(); ++i) {
		balls[i].draw();
	}
	for (int i=0; i<cuboids.size(); ++i) {
		cuboids[i].draw();
	}
	if (!moving && state != WIN_SOLID && state != WIN_STRIPE) {
		cue.draw();
	}
}

// sets the amount of energy that will be put in the stroke
void Game::setStrength(int s) {
	// s = 1 .. 100
	// we want direction to be 0.15 .. 3.0
	float dir = 0.15 + (2.85*(float)s / 100);
	cue.setDirection(cue.getDirection().normalized() * dir);
}

// ball fell in a pocket - remove it from game and change game state
void Game::removeBall(int i) {
	//assert(state == PLAYING_SOLID || state == PLAYING_STRIPE);

	if (balls[i].isBlack()) {	// black ball - either win or lose
		moving = false;
		widget->stop();

		if (state == NEWGAME) {	// if this was the first ball pocketed
			if (setup.countOfSolids + setup.countOfStripes == 0) {	// it also were the only ball in the game -> win
				state = WIN;	
			}
			else {
				state = LOSE;
			}
		}

		if (state == PLAYING_SOLID) {
			if (setup.countOfSolids == 0) {	// solids win
				state = WIN_SOLID;
			}
			else {	// stripe win
				state = WIN_STRIPE;
			}
		}

		else {
			if (setup.countOfStripes == 0) {
				state = WIN_STRIPE;
			}
			else {
				state = WIN_SOLID;
			}
		}
		displayWin();		
	}

	else if (balls[i].isWhite()) {	// cue ball - white
		resetCueBall = true;
	}

	else if (balls[i].isSolid()) {	// solid
		--setup.countOfSolids;
		assert(setup.countOfSolids >= 0);
		if (state == PLAYING_STRIPE) {	// if it was a ball of the other player
			pocketedOther = true;
		}
		else {		// if it was my ball
			pocketedMine = true;
		}
		
		if (state == NEWGAME) {		// first ball pocketed in the game - make solids my color
			state = PLAYING_SOLID;
		}
	}

	else {							// stripe
		assert(balls[i].isStripe());
		--setup.countOfStripes;
		assert(setup.countOfStripes >= 0);
		if (state == PLAYING_SOLID) {	// if it was a ball of the other player
			pocketedOther = true;
		}
		else {						// if it was my ball
			pocketedMine = true;
		}
		
		if (state == NEWGAME) {		// first ball pocketed in the game - make stripes my color
			state = PLAYING_STRIPE;
		}

	}

	// remove the ball from the list of balls - it can't affect the game any more
	balls.erase(balls.begin() + i);
}

// switch whose turn it is in the game
void Game::switchPlayers() {
	if (activePlayer == PLAYING_FIRST) {
		activePlayer = PLAYING_SECOND;
	}
	else {
		activePlayer = PLAYING_FIRST;
	}
	
	if (state != NEWGAME) {		// each player is already assigned what ball type he plays
		if (state == PLAYING_SOLID) {
			state = PLAYING_STRIPE;
		}
		else {
			state = PLAYING_SOLID;
		}
	}
}

// display who won message
void Game::displayWin() {
	string message = "WIN";
	glClearColor(0.0f, 0.3f, 0.0f, 1.0f);

	if (state == WIN_SOLID) {
		message = "Red WINS";
		glClearColor(0.2f, 0.0f, 0.0f, 1.0f);
	}
	else if (state == WIN_STRIPE) {
		message = "Blue WINS";
		glClearColor(0.0f, 0.0f, 0.2f, 1.0f);
	}
	else if (state == LOSE) {
		message = "LOSE";
		glClearColor(0.0f, 0.1f, 0.0f, 1.0f);
	}

	widget->setLabel(message);
	
	widget->updateGL();
}

// display whose turn it is in the game
void Game::displayPlayer() {
	string s = "Player 1";	// so much easier than int->string conversion
	if (activePlayer == PLAYING_SECOND) {
		s = "Player 2";
	}

	if (state != NEWGAME) {		// each player is already assigned what ball type he plays
		if (state == PLAYING_STRIPE) {
			s += " (BLUE)";
		}
		else {
			s += " (RED)";
		}
	}

	widget->setLabel(s);
}

// change view - from "overhead view" to "follow white ball" or vice versa
void Game::changeView() {
	if (overheadView) {
		overheadView = false;
		updateView();	// alter view so that white ball is in the center and looking in the direction of stroke
	}
	else {
		overheadView = true;
		widget->setView(Point(VIEWER_POSITION_X,VIEWER_POSITION_Y,VIEWER_POSITION_Z), Point(0,0,0), Direction(0,1,0));
	}

	widget->updateGL();
}

// alter view so that white ball is in the center and looking in the direction of stroke
void Game::updateView() {
	if (!overheadView) {
		if (!getWhite()) {	// if white ball is not in the game, you can't follow it
			return;
		}
		Point viewer = getWhite()->getCenter() + cue.getDirection().normalized()*(-250) + Direction(0,0,100);
		Point center = getWhite()->getCenter();
		widget->setView(viewer, center, Direction(0,0,1));
	}
}