// sets up the game
// creates the pool table
// positions the balls

#include "GameSetup.h"
#include "Ball.h"
#include "Cuboid.h"
#include "globals.h"
#include <time.h>
#include <fstream>
#include <QColor>

// constructor
GameSetup::GameSetup(const string filename) : setupFile(filename) {
}


// create table and balls for the game
bool GameSetup::readFromFile(vector<Ball> & balls, vector<Cuboid> & cuboids) {
	
	// generate the pool table
	createTable(cuboids);

	countOfSolids = countOfStripes = 0;

	ifstream fin(setupFile.c_str(), fstream::in);	// read only access

	// read balls and other cuboids from a file
	try {	// using file IO can cause exceptions
		string line;

		if (fin.rdstate() & fstream::failbit) {		// if file not found or unable to read from it
			throw fileErrorException("File not found");
		}
		
		float x, y, z, radius, weight, d1x, d1y, d1z, d2x, d2y, d2z, d3x, d3y, d3z;
		int red, green, blue, id, fieldsCount;
		bool whiteFound = false;
		int state = READING_BALLS;

		while (getline(fin, line)) {	// read each line of the file
			if (line == "" || line.at(0) == ';') {	// skip comments and empty lines
				continue;
			}
			
			// directives
			if (line == "#balls") {
				state = READING_BALLS;
				continue;
			}
			else if (line == "#cuboids") {
				state = READING_CUBOIDS;
				continue;
			}
			

			// reading balls from the file
			if (state == READING_BALLS) {
				// values: X Y RADIUS WEIGHT R G B NUMBER
				fieldsCount = sscanf_s(line.c_str(), "%f %f %f %f %d %d %d %d", &x, &y, &radius, &weight, &red, &green, &blue, &id);
				
				if (fieldsCount < 8) {	// ignore the line if it doesn't contain all information
					continue;
				}

				Ball b = Ball(Point(x, y, radius), radius, weight, QColor(red,green,blue), id);
				balls.push_back(b);
			
				if (id == 0) {	// white ball is crucial for the game to work
					whiteFound = true;
				}
				else if (id > 0 && id < 8) {
					++countOfSolids;
				}
				else if (id > 8 && id < 16) {
					++countOfStripes;
				}
				continue;
			}

			// reading cuboids from the file
			else if (state == READING_CUBOIDS) {
				// values: 4*3 float coordinates representing POINT, DIRECTION1, DIRECTION2, DIRECTION3 and 3 integer values - RED GREEN BLUE
				fieldsCount = sscanf_s(line.c_str(), "%f %f %f  %f %f %f  %f %f %f  %f %f %f  %d %d %d", &x, &y, &z, &d1x, &d1y, &d1z, &d2x, &d2y, &d2z, &d3x, &d3y, &d3z, &red, &green, &blue);
				
				if (fieldsCount < 15) {	// ignore the line if it doesn't contain all information
					continue;
				}

				Cuboid c = Cuboid( Point(x,y,z), Direction(d1x, d1y, d1z), Direction(d2x, d2y, d2z), Direction(d3x, d3y, d3z), QColor(red,green,blue) );
				cuboids.push_back(c);
			}

		}
		
		if (!whiteFound) {	// white ball is crucial for the game to work
			throw noWhiteBallException();	// will do a standard setup instead
		}
		fin.close();

	}
	catch (noWhiteBallException) {	// no white ball in the setup file - discard balls read from the file and do a standard setup
		fin.close();
		balls.empty();
		standardSetup(balls, cuboids);
	}
	catch (...) {	// other io errors
		fin.close();
		standardSetup(balls, cuboids);
	}

	return true;
}

// setup implicit game world - pool table and 15 balls in standard position
void GameSetup::standardSetup(vector<Ball>& balls, vector<Cuboid>& cuboids) {
	Cuboid c;
	Point p;
	Direction dir1, dir2, dir3;

	// cue ball - white
	balls.push_back(Ball(Point(WHITE_START_POSITION, RADIUS_WHITE), RADIUS_WHITE, WEIGHT_WHITE, COLOR_WHITE, 0));

	// black ball
	balls.push_back(Ball(Point(BLACKBALL_X, 0, RADIUS), RADIUS, WEIGHT, COLOR_BLACK, 8));
	
	// make the starting positions slightly random
	srand(time(NULL));
	float m = (float)rand() / RAND_MAX / 10;
	float n = (float)rand() / RAND_MAX / 10;
	// solids
	balls.push_back(Ball(Point(BLACKBALL_X-1.9*RADIUS, -RADIUS+n, RADIUS), RADIUS, WEIGHT, COLOR_SOLID, 7));
	balls.push_back(Ball(Point(BLACKBALL_X, 2*RADIUS+m, RADIUS), RADIUS, WEIGHT, COLOR_SOLID, 1));
	balls.push_back(Ball(Point(BLACKBALL_X+1.9*RADIUS, -3*RADIUS-n, RADIUS), RADIUS, WEIGHT, COLOR_SOLID, 6));
	balls.push_back(Ball(Point(BLACKBALL_X+1.9*RADIUS, RADIUS, RADIUS), RADIUS, WEIGHT, COLOR_SOLID, 3));
	balls.push_back(Ball(Point(BLACKBALL_X+3.8*RADIUS+n, -2*RADIUS, RADIUS), RADIUS, WEIGHT, COLOR_SOLID, 2));
	balls.push_back(Ball(Point(BLACKBALL_X+3.8*RADIUS+m, 2*RADIUS, RADIUS), RADIUS, WEIGHT, COLOR_SOLID, 4));
	balls.push_back(Ball(Point(BLACKBALL_X+3.8*RADIUS, 4*RADIUS, RADIUS), RADIUS, WEIGHT, COLOR_SOLID, 5));

	// stripes
	balls.push_back(Ball(Point(BLACKBALL_X-3.8*RADIUS-m, 0, RADIUS), RADIUS, WEIGHT, COLOR_STRIPE, 9));
	balls.push_back(Ball(Point(BLACKBALL_X-1.9*RADIUS, RADIUS, RADIUS), RADIUS, WEIGHT, COLOR_STRIPE, 12));
	balls.push_back(Ball(Point(BLACKBALL_X, -2*RADIUS-n, RADIUS), RADIUS, WEIGHT, COLOR_STRIPE, 15));
	balls.push_back(Ball(Point(BLACKBALL_X+1.9*RADIUS, 3*RADIUS+m, RADIUS), RADIUS, WEIGHT, COLOR_STRIPE, 14));
	balls.push_back(Ball(Point(BLACKBALL_X+1.9*RADIUS, -RADIUS, RADIUS), RADIUS, WEIGHT, COLOR_STRIPE, 10));
	balls.push_back(Ball(Point(BLACKBALL_X+3.8*RADIUS, 0, RADIUS), RADIUS, WEIGHT, COLOR_STRIPE, 13));
	balls.push_back(Ball(Point(BLACKBALL_X+3.8*RADIUS-n, -4*RADIUS-m, RADIUS), RADIUS, WEIGHT, COLOR_STRIPE, 11));

	countOfSolids = 7;
	countOfStripes = 7;
}

// creates pool table (7 cuboids)
void GameSetup::createTable(vector<Cuboid>& cuboids) {
	Cuboid c;
	Point p;
	Direction dir1, dir2, dir3;
	QColor color(0,80,0);

	// POOL TABLE
	// table floor
	p = Point(-(TABLE_LENGTH/2+TABLE_MARGIN), -(TABLE_WIDTH/2+TABLE_MARGIN), -TABLE_MARGIN-1);	// z=-5 to make a safety margin between the bottom and the balls
	dir1 = Direction(TABLE_LENGTH+2*TABLE_MARGIN, 0, 0);
	dir2 = Direction(0, TABLE_WIDTH+2*TABLE_MARGIN, 0);
	dir3 = Direction(0, 0, TABLE_MARGIN);
	c = Cuboid(p, dir1, dir2, dir3, color);
	cuboids.push_back(c);

	color = QColor(10,30,10);
	// left side
	p = Point(-(TABLE_LENGTH/2+TABLE_MARGIN), -TABLE_WIDTH/2+TABLE_POCKET, -1);	// z=-5 to make a safety margin between the bottom and the balls
	dir1 = Direction(TABLE_MARGIN, 0, 0);
	dir2 = Direction(0, TABLE_WIDTH-2*TABLE_POCKET, 0);
	dir3 = Direction(0, 0, TABLE_WALL_HEIGHT);
	c = Cuboid(p, dir1, dir2, dir3, color);
	cuboids.push_back(c);

	// right side
	p = Point(TABLE_LENGTH/2, -TABLE_WIDTH/2+TABLE_POCKET, -1);	// z=-5 to make a safety margin between the bottom and the balls
	dir1 = Direction(TABLE_MARGIN, 0, 0);
	dir2 = Direction(0, TABLE_WIDTH-2*TABLE_POCKET, 0);
	dir3 = Direction(0, 0, TABLE_WALL_HEIGHT);
	c = Cuboid(p, dir1, dir2, dir3, color);
	cuboids.push_back(c);

	// bottom side
	p = Point(-TABLE_LENGTH/2+TABLE_POCKET, -(TABLE_WIDTH/2+TABLE_MARGIN), -1);	// z=-5 to make a safety margin between the bottom and the balls
	dir1 = Direction(TABLE_LENGTH/2-TABLE_POCKET-TABLE_POCKET_MIDDLE/2, 0, 0);
	dir2 = Direction(0, TABLE_MARGIN, 0);
	dir3 = Direction(0, 0, TABLE_WALL_HEIGHT);
	c = Cuboid(p, dir1, dir2, dir3, color);
	cuboids.push_back(c);

	p = Point(TABLE_POCKET_MIDDLE/2, -(TABLE_WIDTH/2+TABLE_MARGIN), -1);	// z=-5 to make a safety margin between the bottom and the balls
	dir1 = Direction(TABLE_LENGTH/2-TABLE_POCKET-TABLE_POCKET_MIDDLE/2, 0, 0);
	dir2 = Direction(0, TABLE_MARGIN, 0);
	dir3 = Direction(0, 0, TABLE_WALL_HEIGHT);
	c = Cuboid(p, dir1, dir2, dir3, color);
	cuboids.push_back(c);

	// top side
	p = Point(-TABLE_LENGTH/2+TABLE_POCKET, TABLE_WIDTH/2, -1);	// z=-5 to make a safety margin between the bottom and the balls
	dir1 = Direction(TABLE_LENGTH/2-TABLE_POCKET-TABLE_POCKET_MIDDLE/2, 0, 0);
	dir2 = Direction(0, TABLE_MARGIN, 0);
	dir3 = Direction(0, 0, TABLE_WALL_HEIGHT);
	c = Cuboid(p, dir1, dir2, dir3, color);
	cuboids.push_back(c);

	p = Point(TABLE_POCKET_MIDDLE/2, TABLE_WIDTH/2, -1);	// z=-5 to make a safety margin between the bottom and the balls
	dir1 = Direction(TABLE_LENGTH/2-TABLE_POCKET-TABLE_POCKET_MIDDLE/2, 0, 0);
	dir2 = Direction(0, TABLE_MARGIN, 0);
	dir3 = Direction(0, 0, TABLE_WALL_HEIGHT);
	c = Cuboid(p, dir1, dir2, dir3, color);
	cuboids.push_back(c);
}
