#include "glwidget.h"
#include "globals.h"
#include "Game.h"
#include "coolaaber.h"
extern Game g;
//extern Coolaaber w;
//class Coolaaber;

/* constructor
 * sets implicit values for variables
 */
GLWidget::GLWidget(QWidget *parent) : QGLWidget(parent) {
	timer = new QTimer(this);
	connect(timer, SIGNAL(timeout()), this, SLOT(timerSlot()));
	g.widget = this;
}

/* this function is automatically called at the beginning of the program
 * initialization of openGL is done here
 */
void GLWidget::initializeGL()
{	
	setView(Point(VIEWER_POSITION_X,VIEWER_POSITION_Y,VIEWER_POSITION_Z), Point(0,0,0), Direction(0,1,0));
	
	// Basic Setup:
	// Set color to use when clearing the background.
	glClearColor(0.0f, 0.0f, 0.0f, 1.0f);
	glClearDepth(1.0f);

	// Turn on backface culling
	glFrontFace(GL_CCW);
	glCullFace(GL_BACK);
	
	// Turn on depth testing
	glEnable(GL_DEPTH_TEST);
	glDepthFunc(GL_LEQUAL);

	// material parameters
	GLfloat mat_specular[]  = {0.5, 0.5, 0.5, 0.9};
	GLfloat mat_shininess[] = {5.0};
	glShadeModel(GL_SMOOTH);
	glMaterialfv(GL_FRONT, GL_SPECULAR, mat_specular);
	glMaterialfv(GL_FRONT, GL_SHININESS, mat_shininess);

	glEnable(GL_NORMALIZE);
	glEnable(GL_COLOR_MATERIAL);
	glEnable(GL_LIGHTING);
	glEnable( GL_POLYGON_SMOOTH );
	
	/* Add lights to the scene */
	GLfloat lightPosition[] = {VIEWER_POSITION_X, VIEWER_POSITION_Y, VIEWER_POSITION_Z, 10};
	GLfloat white_light[] = {0.9, 0.9, 0.9, 0.9};
	GLfloat black_light[] = {0.1, 0.1, 0.1, 0.5};
	glLightfv( GL_LIGHT1, GL_POSITION, lightPosition );
	glLightfv( GL_LIGHT1, GL_AMBIENT, white_light );
	glLightfv( GL_LIGHT1, GL_DIFFUSE, white_light );
	glLightfv( GL_LIGHT1, GL_SPECULAR, white_light );
	glLightf(GL_LIGHT1, GL_CONSTANT_ATTENUATION, 2);
	glEnable(GL_LIGHT1);

	// add fog effect
	GLfloat fogColor[4] = { 10.f/255, 30.f/255, 30.f/255, 10 };
	glFogi( GL_FOG_MODE, GL_EXP2 );
	glFogfv( GL_FOG_COLOR, fogColor );
	glFogf( GL_FOG_DENSITY, 0.1 );
	glEnable( GL_FOG );
}

/* this function is automatically called after each repaint event
 * it works as display function in pure openGL
 */
void GLWidget::paintGL()
{
	glMatrixMode( GL_MODELVIEW );
	glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT ); // erase buffer with colors = erase screen
	g.draw();
}

/* this function is automatically called after a mouse button was pressed */
void GLWidget::mousePressEvent( QMouseEvent *event )
{
	setCursor( Qt::BlankCursor );
	lastPosition = Point(event->x(), event->y(), 0);		// remember last position
}

/* this function is automatically called after a mouse button was released */
void GLWidget::mouseReleaseEvent( QMouseEvent *event )
{
	// do a stroke
	if (event->button() == Qt::RightButton) {
		g.stroke();
	}
	unsetCursor();
}

/* this function is automatically called after mouse has been moved while a button is being pressed */
void GLWidget::mouseMoveEvent( QMouseEvent *event )
{
	// rotate cue direction - aiming
	//if ( event->button() == Qt::LeftButton ) {
		g.rotateCue((float)(event->x()-lastPosition.getX()) / MOUSE_SENSITIVITY, 0);
		lastPosition = Point(event->x(), event->y(), 0);
		updateGL();
	//}
}

// start the animation timer
void GLWidget::start(int interval) {
	timer->start(interval);
}

// stop the timer
void GLWidget::stop() {
	timer->stop();
}

/* this function is automatically called after the openGL widget has been resized */
void GLWidget::resizeGL(int width, int height)
{
	int w = height * WIDTH/HEIGHT;

	glViewport((width-w)/2,0, w, height);
}

/* slot that is called after each timer's time-out
 * this is the core of animation
 */
void GLWidget::timerSlot() {
	g.doIteration();	// do simulation iteration (actually more of them - based on setup)
	if (displayEnabled) {
		updateGL();
	}
}

// stroke energy - move event
void GLWidget::strengthValueSlot(int s) {
	strength = s+1;
}

// stroke energy - slider release event
void GLWidget::strengthSlot() {
	g.setStrength(strength);
}

// set up access to other parts of the application
void GLWidget::coolClass(Coolaaber* c) {
	cool = c;
}

// restart to new game
void GLWidget::restartSlot() {
	g.newGame();
	g.displayPlayer();
	cool->ui.strengthSlider->setValue(75);
	updateGL();
}

// display message for the user
void GLWidget::setLabel(const std::string& s) {
	cool->ui.playerLabel->setText(s.c_str());
}

// change view
// white ball in the center, looking in the direction of stroke
void GLWidget::setView(const Point& viewer, const Point& center, const Direction& up) {
	glMatrixMode( GL_PROJECTION );
	glLoadIdentity();
	gluPerspective(30, WIDTH/HEIGHT, 100, 700);
	gluLookAt(viewer.getX(), viewer.getY(), viewer.getZ(), center.getX(), center.getY(), center.getZ(), up.getX(), up.getY(), up.getZ());
}

// change view button slot
void GLWidget::changeViewSlot() {
	g.changeView();
}